import hashlib
from datetime import datetime, timedelta

from django.urls import reverse
from jwcrypto.jwt import JWT

from django_jwt.server.models import Key
from django_jwt.settings_utils import get_setting
from django_jwt.view_utils import calculate_at_hash


def get_value(user, names):
    obj = user
    for name in names:
        obj = getattr(obj, name, None)
        if callable(obj):
            obj = obj()
    return obj


def create_jwt_code(request, session, token, access_token=None):
    now = datetime.now()
    expiration_time = timedelta(seconds=get_setting('JWT_SERVER.JWT_EXPIRATION_TIME'))
    jwk = Key.get_actual_jwk()
    if token == 'access_token':
        claims = {'aud': [request.build_absolute_uri(reverse('userinfo_endpoint'))],
                  'azp': session.web.id, 'scope': 'openid profile', 'sub': request.user.pk}
        expiration_time = timedelta(seconds=7200)
    else:
        claims = {'aud': [session.web.id]}
        if access_token is not None:
            claims['at_hash'] = calculate_at_hash(access_token, hashlib.sha256)
        for attribute in session.web.attributewebpage_set.filter(restrict=False):
            claims[attribute.attribute] = get_value(request.user, attribute.value.split('.'))
        if claims.get('sub', None) is None:
            claims['sub'] = request.user.pk
    claims.update({'iss': request.build_absolute_uri('/'),
                   'iat': int(now.timestamp()),
                   'exp': int((now + expiration_time).timestamp()),
                   'nonce': request.GET.get('nonce')})
    jwt = JWT(header={'kid': jwk.kid, 'alg': 'RS256', 'typ': 'JWT'},
              claims=claims)
    jwt.make_signed_token(jwk)
    return jwt.serialize()



from django.contrib.auth.models import AnonymousUser
from django.utils.deprecation import MiddlewareMixin

from django_jwt.auth import JWTAuthentication
from django_jwt.settings_utils import get_setting


class JWTAuthenticationMiddleware(MiddlewareMixin):
    def process_request(self, request):
        key = request.get_signed_cookie(get_setting('JWT_CLIENT.COOKIE_NAME'), None, salt=get_setting('SECRET_KEY'))
        user = None
        if key is not None:
            try:
                user = JWTAuthentication.authenticate_credentials(key)
            except JWTAuthentication.JWTException:
                pass
        request.user = user or AnonymousUser()



from django.conf import settings


class NoDefault:
    pass


JWT_DEFAULT_SETTINGS = {
    'JWT_CLIENT.RENAME_ATTRIBUTES': {},
    'JWT_CLIENT.DEFAULT_ATTRIBUTES': {},
    'JWT_CLIENT.CREATE_USER': False,
    'JWT_CLIENT.COOKIE_NAME': 'id_token',
    'JWT_SERVER.JWK_EXPIRATION_TIME': 3600,
    'JWT_SERVER.JWT_EXPIRATION_TIME': 14400,
    'LOGOUT_URL': 'logout',
    'JWT_CLIENT.RESPONSE_TYPE': '#',
}


def convert_url(url):
    if len(url) > 0 and url[-1] == '/':
        url = url[:-1]
    return url


JWT_DEFAULT_CONVERTERS = {
    'JWT_CLIENT.OPENID2_URL': convert_url,
}


def get_setting(names):
    default = JWT_DEFAULT_SETTINGS.get(names, NoDefault)
    value = settings
    for name in names.split('.'):
        if isinstance(value, dict):
            value = value.get(name, default)
        else:
            value = getattr(value, name, default)
    if value == NoDefault:
        raise Exception('Setting %s is required' % names)
    converter = JWT_DEFAULT_CONVERTERS.get(names, lambda x: x)
    value = converter(value)
    return value


def get_domain_from_url(url):
    return '/'.join(url.split('/')[0:3])



from django import forms

from django_jwt.server.models import AttributeWebPage, WebPage


class IdTokenExtraClaimAdminForm(forms.ModelForm):
    restrict = forms.BooleanField(widget=forms.HiddenInput(), initial=False, required=False)

    class Meta:
        model = AttributeWebPage
        exclude = ()
        labels = {
            'attribute': 'Claim name',
            'value': 'User attribute',
        }
        help_texts = {
            'value': 'You can use methods without "()" and use "." to get an attribute inside another one.'
        }


class RestrictUsersAdminForm(forms.ModelForm):
    restrict = forms.BooleanField(widget=forms.HiddenInput(), initial=True, required=False)

    class Meta:
        model = AttributeWebPage
        exclude = ()
        labels = {
            'attribute': 'User attribute',
        }
        help_texts = {
            'value': 'If the user has the same value as this, it won\'t log in into this web page.'
        }


class WebPageAdminForm(forms.ModelForm):
    class Meta:
        model = WebPage
        exclude = ()
        labels = {
            'logout_all': 'Log out from main page',
            'response_type': 'Log in response type'
        }
        help_texts = {
            'host': 'Web page url without the last "/". Example: https://localhost:8000',
            'response_type': 'Hash response is for frontend auth and Params for backend',
        }



from django.urls import path

from django_jwt import views
from django_jwt.settings_utils import get_setting

urlpatterns = [
    path('login', views.LoginView.as_view(), name='oidc_login'),
    path('logout', views.LogoutView.as_view(), name='oidc_logout'),
]

if get_setting('JWT_CLIENT.TYPE') == 'fake':
    urlpatterns.extend([
        path('jwks', views.jwks, name="fake_jwks"),
        path('fake-login', views.fake_login, name="fake_login"),
        path('.well-known/openid-configuration', views.fake_config, name='fake_config'),
        path('fake-userinfo', views.fake_userinfo, name='fake_userinfo'),
    ])



import hashlib
import json
import os.path

import base64
from datetime import datetime, timedelta

from Crypto.PublicKey import RSA
from jwcrypto.jwk import JWK
from jwcrypto.jwt import JWT

from django_jwt.patterns import Singleton
from django_jwt.settings_utils import get_setting


class FakeJWT(metaclass=Singleton):
    def __init__(self) -> None:
        super().__init__()
        self.jks = self.get_jwk()

    def get_jwk(self):
        if not os.path.isdir('files'):
            os.mkdir('files')
        if not os.path.isfile('files/rsa.pem'):
            with open('files/rsa.pem', 'wb+') as file:
                private_key = RSA.generate(2048)
                private_key = private_key.export_key()
                file.write(private_key)
        else:
            with open('files/rsa.pem', 'rb') as file:
                private_key = file.read()
        jwk = JWK()
        jwk.import_from_pem(private_key, kid='0')
        return jwk

    def generate_jwt(self, claim):
        jwt = JWT(header={'kid': self.jks.key_id, 'alg': 'RS256', 'typ': 'JWT'}, claims=claim)
        jwt.make_signed_token(self.jks)
        return jwt.serialize()


def get_jwks():
    return {'keys': [FakeJWT().jks.export_public(as_dict=True)]}


def get_sub_jwt(token):
    jwt = JWT()
    jwt.deserialize(token, FakeJWT().jks)
    claims = json.loads(jwt.claims)
    return claims.get('sub')


def calculate_at_hash(access_token, hash_alg):
    """Helper method for calculating an access token
    hash, as described in http://openid.net/specs/openid-connect-core-1_0.html#CodeIDToken
    Its value is the base64url encoding of the left-most half of the hash of the octets
    of the ASCII representation of the access_token value, where the hash algorithm
    used is the hash algorithm used in the alg Header Parameter of the ID Token's JOSE
    Header. For instance, if the alg is RS256, hash the access_token value with SHA-256,
    then take the left-most 128 bits and base64url encode them. The at_hash value is a
    case sensitive string.
    Args:
        access_token (str): An access token string.
        hash_alg (callable): A callable returning a hash object, e.g. hashlib.sha256
    """
    hash_digest = hash_alg(access_token.encode('ascii')).digest()
    cut_at = int(len(hash_digest) / 2)
    truncated = hash_digest[:cut_at]
    at_hash = base64url_encode(truncated)
    return at_hash.decode('ascii')


def base64url_encode(input):
    """Helper method to base64url_encode a string.
    Args:
        input (str): A base64url_encoded string to encode.
    """
    return base64.urlsafe_b64encode(input).replace(b'=', b'')


def crear_url_amb_jwt(request):
    fake_jwt = FakeJWT()
    now = datetime.now()
    expiration = timedelta(days=1)
    claim = {'sub': request.POST['username'], 'nonce': request.GET.get('nonce'), 'iat': int(now.timestamp()),
             'exp': int((now + expiration).timestamp()), 'iss': request.build_absolute_uri('/'),
             'aud': [request.GET.get('client_id')]}
    access_token = fake_jwt.generate_jwt(claim=claim)
    claim['at_hash'] = calculate_at_hash(access_token, hashlib.sha256)
    id_token = fake_jwt.generate_jwt(claim=claim)
    url = "%s%saccess_token=%s&id_token=%s&state=%s" % (request.GET.get('redirect_uri'),
                                                        get_setting('JWT_CLIENT.RESPONSE_TYPE'), access_token,
                                                        str(id_token), request.GET.get('state'))
    return url



import json

from django.contrib.auth import get_user_model
from django.contrib.auth.mixins import LoginRequiredMixin
from django.core.exceptions import PermissionDenied, SuspiciousOperation
from django.http import JsonResponse
from django.shortcuts import redirect, render
from django.urls import reverse
from django.utils.http import urlencode
from django.views import View
from django.views.generic import TemplateView
from django_jwt.auth import JWTAuthentication

from django_jwt.settings_utils import get_setting

from django_jwt.server.models import WebPage, UserWebPagePermission, UserExternalSession, Key, AttributeWebPage
from django_jwt.server.views_utils import create_jwt_code, get_value


class OpenIdConfiguration(View):
    def get(self, request):
        configuration = {
            'issuer': request.build_absolute_uri('/'),
            'authorization_endpoint': request.build_absolute_uri(reverse('authorization_endpoint')),
            # 'userinfo_endpoint': request.build_absolute_uri(reverse('userinfo_endpoint')),
            'jwks_uri': request.build_absolute_uri(reverse('jwks_uri')),
            'end_session_endpoint': request.build_absolute_uri(reverse('end_session_endpoint')),
            'response_types_supported': ['id_token', 'token'],
            'subject_types_supported': ['public'],
            'id_token_signing_alg_values_supported': ['RS256'],
            'claims_supported': ['sub', 'iss', 'aud', 'exp', 'iat', 'jti', 'scope', 'azp'],
            'require_request_uri_registration': True,
        }
        return JsonResponse(configuration)


class AuthorizationView(LoginRequiredMixin, TemplateView):
    template_name = 'oidc_confirmation.html'

    def get_login_url(self):
        connection = self.request.GET.get('connection', '')
        return super().get_login_url() + (('-' + connection) if connection else '')

    def dispatch(self, request, *args, **kwargs):
        try:
            kwargs['web'] = WebPage.objects.prefetch_related('attributewebpage_set')\
                .get(id=self.request.GET.get('client_id'))
        except WebPage.DoesNotExist:
            raise PermissionDenied()
        redirect_url = self.request.GET.get('redirect_uri', None)
        if redirect_url is None or not redirect_url.startswith(kwargs['web'].host + '/'):
            raise PermissionDenied()
        return super().dispatch(request, *args, **kwargs)

    def get(self, request, *args, **kwargs):
        web = kwargs.get('web')
        self.check_user_permissions(web)
        if web.needs_confirmation:
            try:
                UserWebPagePermission.objects.get(web=web, user=request.user)
            except UserWebPagePermission.DoesNotExist:
                return render(request, self.template_name, context={'web': web})
        return self.success_response(web)

    def success_response(self, web):
        external_session = UserExternalSession.objects\
            .get_or_create(web=web, session_id=self.request.session.session_key)[0]
        tokens = self.request.GET.get('response_type', None)
        if tokens is None:
            raise SuspiciousOperation("Invalid request; see documentation for correct paramaters")
        tokens = tokens.split(' ')
        response_parameters = {'state': self.request.GET.get('state')}
        if 'token' in tokens:
            response_parameters['access_token'] = create_jwt_code(request=self.request, token='access_token',
                                                                  session=external_session)
        if 'id_token' in tokens:
            response_parameters['id_token'] = create_jwt_code(
                request=self.request, token='id_token', session=external_session,
                access_token=response_parameters.get('access_token', None))
        redirect_uri = self.request.GET.get('redirect_uri')
        if web.response_type in redirect_uri:
            redirect_uri += '&' + urlencode(response_parameters)
        else:
            redirect_uri += web.response_type + urlencode(response_parameters)
        return redirect(redirect_uri)

    def post(self, request, *args, **kwargs):
        web = kwargs.get('web')
        if request.POST.get('confirmation', 'false') == 'true':
            UserWebPagePermission(web=web, user=request.user).save()
            return redirect(request.get_full_path())
        return render(request, self.template_name, context={'web': web, 'denied': True})

    def check_user_permissions(self, web):
        for restrict in web.attributewebpage_set.filter(restrict=True):
            if str(get_value(self.request.user, restrict.attribute.split('.'))) == restrict.value:
                raise PermissionDenied('You don\'t have permission to log in into this web page.')


class JWKsView(View):
    def get(self, request):
        return JsonResponse({'keys': Key.get_jwk_set()})


class LogoutView(View):
    def dispatch(self, request, *args, **kwargs):
        token = request.GET.get('id_token_hint', None)
        redirect_url = self.request.GET.get('post_logout_redirect_uri', '')
        host = '/'.join(redirect_url.split('/')[:3])
        if token is None:
            raise PermissionDenied()
        try:
            kwargs['web'] = WebPage.objects.get(host=host)
        except WebPage.DoesNotExist:
            raise PermissionDenied()
        kwargs['user'] = self.get_user(token, kwargs['web'])
        return super().dispatch(request, *args, **kwargs)

    def get_user(self, token, web):
        try:
            jwt = JWTAuthentication.validate_jwt(token)
        except JWTAuthentication.JWTException:
            raise PermissionDenied()
        claims = json.loads(jwt.claims)
        if not self.check_aud(claims, web):
            raise PermissionDenied()
        if not JWTAuthentication.verify_claims(claims, client_id=web.id):
            raise PermissionDenied()
        try:
            att = web.attributewebpage_set.get(attribute='sub').value.replace('.', '__')
        except AttributeWebPage.DoesNotExist:
            att = 'id'
        kwargs = {att: claims.get('sub')}
        User = get_user_model()
        try:
            return User.objects.get(**kwargs)
        except User.DoesNotExist:
            raise PermissionDenied()

    def check_aud(self, claims, web):
        if isinstance(claims['aud'], str):
            claims['aud'] = [claims['aud']]
        for aud in claims['aud']:
            if aud in [web.host, web.id]:
                return True
        return False

    def get(self, request, *args, **kwargs):
        if kwargs['web'].logout_all and request.user.is_authenticated:
            if kwargs['user'].pk != request.user.pk:
                raise PermissionDenied()
            params = {'next': request.build_absolute_uri()}
            return redirect(reverse(get_setting('LOGOUT_URL')) + '?' + urlencode(params))
        if not kwargs['web'].logout_all:
            UserExternalSession.objects.filter(web=kwargs['web'], session_id=request.session.session_key).delete()
        return redirect(request.GET.get('post_logout_redirect_uri'))



# Generated by Django 4.0 on 2022-03-03 11:35

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('server', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='webpage',
            name='logout_all',
            field=models.BooleanField(default=True),
        ),
        migrations.AddField(
            model_name='webpage',
            name='response_type',
            field=models.CharField(choices=[('#', 'Hash'), ('?', 'Params')], default='#', max_length=1),
        ),
    ]



import json
import logging

import urllib3
from jwcrypto.jwk import JWKSet

from django_jwt.exceptions import JWTClientException
from django_jwt.patterns import Singleton
from django_jwt.settings_utils import get_setting


# singleton class that saves all information from the OpenId server
class OpenId2Info(metaclass=Singleton):
    def __init__(self):
        url = get_setting('JWT_CLIENT.OPENID2_URL')
        client_type = get_setting('JWT_CLIENT.TYPE')
        if client_type == 'local':
            apps = get_setting('INSTALLED_APPS')
            if 'django_jwt.server' not in apps:
                raise JWTClientException('django_jwt.server not installed for type local')
        url += '/.well-known/openid-configuration'

        # Getting urls info from OpenId server
        http = urllib3.PoolManager()
        r = http.request('GET', url)
        if r.status != 200:
            error_message = 'OpenID returned error code %s on openid-configuration: %s' % (r.status, url)
            logger = logging.getLogger(__name__)
            logger.critical(error_message)
            raise JWTClientException(error_message)
        data = json.loads(r.data.decode('UTF-8'))
        self.jwks_uri = data.get('jwks_uri', None)
        self.authorization_endpoint = data.get('authorization_endpoint', None)
        self.end_session_endpoint = data.get('end_session_endpoint', None)
        if self.jwks_uri is None or self.authorization_endpoint is None:
            raise JWTClientException('jkws_uri or authorization_endpoint not found in OpenId openid-configuration url')
        self.fetch_jwks()

    # Updates all the actual JWKS from the OPenId server
    def fetch_jwks(self):
        logger = logging.getLogger(__name__)
        logger.info('Fetching JWKs')
        http = urllib3.PoolManager()
        r = http.request('GET', self.jwks_uri)
        if r.status != 200:
            error_message = 'OpenID returned error code %s on jwks_uri: %s' % (r.status, self.jwks_uri)
            logger.critical(error_message)
            raise JWTClientException(error_message)
        self.jwks = JWKSet.from_json(r.data.decode('UTF-8'))



import logging

from rest_framework import exceptions
from rest_framework.authentication import TokenAuthentication
from django.utils.translation import gettext_lazy as _

from django_jwt.auth import JWTAuthentication


class JWTTokenAuthentication(TokenAuthentication):
    keyword = 'Bearer'

    def authenticate_credentials(self, key):
        try:
            user = JWTAuthentication.authenticate_credentials(key)
        except JWTAuthentication.JWTException as e:
            logger = logging.getLogger(__name__)
            logger.warning(e)
            raise exceptions.AuthenticationFailed(_('Token is invalid or expired.'))
        if not user:
            raise exceptions.AuthenticationFailed(_('User does not exist.'))
        return user, key



from django.contrib import admin

from django_jwt.server.forms import IdTokenExtraClaimAdminForm, RestrictUsersAdminForm, WebPageAdminForm
from django_jwt.server.models import WebPage, AttributeWebPage


class WebPagesAttributesAdmin(admin.StackedInline):
    verbose_name = 'ID Token extra claim'
    model = AttributeWebPage
    extra = 1
    form = IdTokenExtraClaimAdminForm

    def get_queryset(self, request):
        return super().get_queryset(request).filter(restrict=False)


class RestrictUsersAdmin(admin.StackedInline):
    verbose_name = 'User attribute restricted'
    verbose_name_plural = 'User attributes restricted'
    model = AttributeWebPage
    extra = 1
    form = RestrictUsersAdminForm

    def get_queryset(self, request):
        return super().get_queryset(request).filter(restrict=True)


@admin.register(WebPage)
class WebPageFullAdmin(admin.ModelAdmin):
    readonly_fields = ('id',)
    inlines = [WebPagesAttributesAdmin, RestrictUsersAdmin]
    form = WebPageAdminForm



from django.core.exceptions import PermissionDenied
from django.http import JsonResponse, HttpResponseBadRequest
from django.shortcuts import render, redirect
from django.urls import reverse
from django.utils.crypto import get_random_string
from django.utils.http import urlencode
from django.views import View

from django_jwt.auth import JWTAuthentication
from django_jwt.openid import OpenId2Info
from django_jwt.settings_utils import get_setting
from django_jwt.view_utils import crear_url_amb_jwt, get_jwks, get_sub_jwt


class LoginView(View):
    def get(self, request, *args, **kwargs):
        id_token = request.GET.get('id_token', None)
        if id_token is None:
            request.session['state'] = get_random_string(16)
            request.session['nonce'] = get_random_string(32)
            params = {
                'client_id': get_setting('JWT_CLIENT.CLIENT_ID'),
                'state': request.session['state'],
                'nonce': request.session['nonce'],
                'redirect_uri': request.build_absolute_uri(),
                'response_type': get_setting('JWT_CLIENT.RESPONSE_TYPE')
            }
            return redirect(OpenId2Info().authorization_endpoint + '?' + urlencode(params))
        req_state = request.GET.get('state', None)
        if req_state is not None and req_state != request.session.get('state', None):
            return HttpResponseBadRequest()
        try:
            user = JWTAuthentication.authenticate_credentials(id_token, nonce=request.session.get('nonce', ''))
        except JWTAuthentication.JWTException:
            return HttpResponseBadRequest()
        if user is None:
            raise PermissionDenied()
        response = redirect(request.GET.get('next', '/'))
        response.set_signed_cookie(get_setting('JWT_CLIENT.COOKIE_NAME'), id_token, salt=get_setting('SECRET_KEY'))
        return response


class LogoutView(View):
    def get(self, request, *args, **kwargs):
        redirect_uri = get_setting('LOGOUT_REDIRECT_URL')
        if request.user.is_authenticated and OpenId2Info().end_session_endpoint is not None:
            token = request.get_signed_cookie(get_setting('JWT_CLIENT.COOKIE_NAME'), None,
                                              salt=get_setting('SECRET_KEY'))
            params = {'post_logout_redirect_uri': request.build_absolute_uri(), 'id_token_hint': token}
            response = redirect(OpenId2Info().end_session_endpoint + '?' + urlencode(params))
            response.delete_cookie(get_setting('JWT_CLIENT.COOKIE_NAME'))
        else:
            response = redirect(redirect_uri)
        response.delete_cookie(get_setting('JWT_CLIENT.COOKIE_NAME'))
        return response


def fake_config(request):
    config = {
        'issuer': request.build_absolute_uri('/'),
        'userinfo_endpoint': request.build_absolute_uri(reverse('fake_userinfo')),
        'authorization_endpoint': request.build_absolute_uri(reverse('fake_login')),
        'jwks_uri': request.build_absolute_uri(reverse('fake_jwks')),
        'response_types_supported': ['id_token'],
        'subject_types_supported': ['public'],
        'id_token_signing_alg_values_supported': ['RS256'],
        'claims_supported': ['sub', 'iss', 'aud', 'exp', 'iat', 'jti', 'scope', 'azp'],
        'require_request_uri_registration': True,
    }
    return JsonResponse(config)


def fake_login(request):
    if request.method == 'GET':
        return render(request, "fake_login.html")
    if request.method == 'POST':
        return redirect(crear_url_amb_jwt(request))


def jwks(request):
    return JsonResponse(get_jwks())


def fake_userinfo(request):
    sub = get_sub_jwt(request.headers.get('Authorization').split(' ')[1])
    return JsonResponse({'sub': sub})



import os

import setuptools

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, 'README.md')).read()

setuptools.setup(
    name='django-jwt-oidc',
    version=os.getenv('PACKAGE_VERSION').split('/')[-1],
    packages=setuptools.find_packages(),
    include_package_data=True,
    description='Django library that implements the authentification for OpenId SSO with JWT from oauth2.',
    long_description=README,
    long_description_content_type="text/markdown",
    author='Arnau Casas Saez',
    author_email='casassarnau@gmail.com',  # SEE NOTE BELOW (*)
    url='https://github.com/Casassarnau/django-jwt',
    license='MIT',
    python_requires='>=3.6',
    project_urls={
        "Bug Tracker": "https://github.com/Casassarnau/django-jwt/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'django>=2.2',
        'pycryptodome',
        'jwcrypto',
        'urllib3'
    ]
)

# (*) Please direct queries to the discussion group, rather than to me directly
#     Doing so helps ensure your question is helpful to other users.
#     Queries directly to my email are likely to receive a canned response.
#
#     Many thanks for your understanding.
